/**
 * \file backtrace_gcc.cpp
 * Definition of the backtrace for GCC compiler, or more precisely for the platforms that support glib
 */

#include "backtrace.h"

#include <fstream>
#include <list>
#include <string>

#include <execinfo.h>
#include <signal.h>

/// Max depth for the call stack
const unsigned int callStackDepth = 50;

/// The real implementation of the backtrace dump
struct BacktraceImpl
{
  /// Name of the file in which the dump will be made
  static std::string string;
  /// The call stack elements
  std::list<std::string> callStack;

  /// Populate the callStack string list
  void populateCallStack()
  {
    void *array[callStackDepth];
    size_t size;
    char **strings;
    size = backtrace(array, callStackDepth);
    strings = backtrace_symbols(array, size);

    for(size_t i = 0; i < size; ++i)
    {
      callStack.push_back(strings[i]);
    }
    free(strings);
  }

  /**
   * Dumps the call stack list on a stream
   * @param stream is the stream on which the call stack will be dumped
   */
  void dumpCallStack(std::ostream& stream) const
  {
    stream << "**\nCall Stack\n";
    for(std::list<std::string>::const_iterator it = callStack.begin(); it != callStack.end(); ++it)
    {
      stream << *it << std::endl;
    }
    stream << "**" << std::endl;
  }
};

/**
 * Handles the unhandled exception
 * @param signal is the number of the signal that triggered the function
 */
void UnhandledExceptionFilter_int(int signal)
{
  BacktraceImpl impl;
  impl.populateCallStack();
  std::ofstream stream(BacktraceImpl::string.c_str());
  switch(signal)
  {
    case SIGINT:
      stream << "Terminal interruption signal\n";
      break;
    case SIGQUIT:
      stream << "Quit\n";
      break;
    case SIGILL:
      stream << "Illegal instruction\n";
      break;
    case SIGABRT:
      stream << "Abort\n";
      break;
    case SIGFPE:
      stream << "Floating point exception\n";
      break;
    case SIGKILL:
      stream << "Killed\n";
      break;
    case SIGSEGV:
      stream << "Segmentation fault\n";
      break;
    case SIGPIPE:
      stream << "Pipe error\n";
      break;
    case SIGALRM:
      stream << "Alarm clock\n";
      break;
    case SIGTERM:
      stream << "Terminated\n";
      break;
    default:
      stream << "Unknown unhandled exception number " << signal << "\n";
  }
  impl.dumpCallStack(stream);
  exit(-1);
}

void UnhandledExceptionFilter()
{
  UnhandledExceptionFilter_int(-1);
}

void setExceptionHandler(const std::string& string)
{
  BacktraceImpl::string = string;
  signal(SIGINT,  UnhandledExceptionFilter_int);
  signal(SIGQUIT,  UnhandledExceptionFilter_int);
  signal(SIGILL,  UnhandledExceptionFilter_int);
  signal(SIGABRT,  UnhandledExceptionFilter_int);
  signal(SIGFPE,  UnhandledExceptionFilter_int);
  signal(SIGKILL,  UnhandledExceptionFilter_int);
  signal(SIGSEGV,  UnhandledExceptionFilter_int);
  signal(SIGPIPE,  UnhandledExceptionFilter_int);
  signal(SIGALRM,  UnhandledExceptionFilter_int);
  signal(SIGTERM,  UnhandledExceptionFilter_int);
}

std::string BacktraceImpl::string;
