/**
 * \file backtrace_msvc.cpp
 * Definition of the backtrace for MSVC compiler, or more precisely for the platforms that support dbghelp
 */

#include "backtrace.h"
#include <list>
#include <string>
#include <fstream>

#include <boost/lexical_cast.hpp>

#include <windows.h>
#include <dbghelp.h>

const size_t nbChar = 100;

BOOL CALLBACK EnumerateLoadedModules(PTSTR ModuleName, DWORD64 ModuleBase, ULONG ModuleSize, PVOID UserContext)
{
  std::list<std::string>* pLoadedModules = reinterpret_cast<std::list<std::string>*>(UserContext);

  pLoadedModules->push_back(std::string(ModuleName) + "\t" + boost::lexical_cast<std::string>(ModuleBase));
  return true;
}

struct BacktraceImpl
{
  /// Name of the file in which the dump will be made
  static std::string string;
  /// Handle to the process
  HANDLE hProcess;

  /// List of the loaded modules
  std::list<std::string> loadedModules;
  /// List of the elements on the call stack
  std::list<std::string> callStack;
  /// List of the elements on the file stack
  std::list<std::string> fileStack;

  /**
   * Dumps the loaded modules on the stream
   * @param stream is the stream to which the dump will be made
   */
  void dumpLoadedModules(std::ostream& stream)
  {
    stream << "**\nLoaded Modules\n";
    for(std::list<std::string>::const_iterator it = loadedModules.begin(); it != loadedModules.end(); ++it)
    {
      stream << "- " << *it << std::endl;
    }
    stream << "**" << std::endl;
  }

  /**
   * Loads the elements of the call stack in a list
   * @param exceptionInfos are useful information on the exception
   */
  void LoadCallStack(EXCEPTION_POINTERS* exceptionInfos)
  {
    STACKFRAME64 tempStackFrame;
    CONTEXT context = *(exceptionInfos->ContextRecord);
    memset( &tempStackFrame, 0, sizeof(STACKFRAME64) );
    DWORD machineType;

#ifdef _M_IX86
    machineType = IMAGE_FILE_MACHINE_I386;
    tempStackFrame.AddrPC.Offset       = context.Eip;
    tempStackFrame.AddrPC.Mode         = AddrModeFlat;
    tempStackFrame.AddrStack.Offset    = context.Esp;
    tempStackFrame.AddrStack.Mode      = AddrModeFlat;
    tempStackFrame.AddrFrame.Offset    = context.Ebp;
    tempStackFrame.AddrFrame.Mode      = AddrModeFlat;
#elif _M_X64
    machineType = IMAGE_FILE_MACHINE_AMD64;
    tempStackFrame.AddrPC.Offset = context.Rip;
    tempStackFrame.AddrPC.Mode = AddrModeFlat;
    tempStackFrame.AddrFrame.Offset = context.Rsp;
    tempStackFrame.AddrFrame.Mode = AddrModeFlat;
    tempStackFrame.AddrStack.Offset = context.Rsp;
    tempStackFrame.AddrStack.Mode = AddrModeFlat;
#elif _M_IA64
    machineType = IMAGE_FILE_MACHINE_IA64;
    tempStackFrame.AddrPC.Offset = context.StIIP;
    tempStackFrame.AddrPC.Mode = AddrModeFlat;
    tempStackFrame.AddrFrame.Offset = context.IntSp;
    tempStackFrame.AddrFrame.Mode = AddrModeFlat;
    tempStackFrame.AddrBStore.Offset = context.RsBSP;
    tempStackFrame.AddrBStore.Mode = AddrModeFlat;
    tempStackFrame.AddrStack.Offset = context.IntSp;
    tempStackFrame.AddrStack.Mode = AddrModeFlat;
#else
#error "Platform not supported!"
#endif

    ULONG64 buffer[(sizeof(SYMBOL_INFO) + nbChar*sizeof(TCHAR) + sizeof(ULONG64) + 1) / sizeof(ULONG64)];
    PSYMBOL_INFO pSymbol = reinterpret_cast<PSYMBOL_INFO>(buffer);
    PTSTR undecoratedName = (PTSTR)malloc(sizeof(TCHAR) * nbChar);

    pSymbol->SizeOfStruct = sizeof(SYMBOL_INFO);
    pSymbol->MaxNameLen = nbChar;
    DWORD lineDisplacement;
    IMAGEHLP_LINE64 lineInfo = { sizeof(IMAGEHLP_LINE64) };

    while(StackWalk64(machineType, hProcess, GetCurrentThread(), &tempStackFrame, &context, NULL, SymFunctionTableAccess64, SymGetModuleBase64, NULL))
    {
      // Sanity stack check
      if(tempStackFrame.AddrPC.Offset == 0)
        break;

      DWORD64 symDisplacement = 0;
      // Try to get the symbol name
      if(SymFromAddr(hProcess, tempStackFrame.AddrPC.Offset, &symDisplacement, pSymbol))
      {
        UnDecorateSymbolName(pSymbol->Name, undecoratedName, MAX_SYM_NAME, UNDNAME_COMPLETE);
        callStack.push_back(std::string(undecoratedName) + "+" + boost::lexical_cast<std::string>(symDisplacement));

        if(SymGetLineFromAddr64(hProcess, tempStackFrame.AddrPC.Offset, &lineDisplacement, &lineInfo))
        {
          fileStack.push_back(std::string(lineInfo.FileName) + "\tl:" + boost::lexical_cast<std::string>(lineInfo.LineNumber));
        }
        else
        {
          fileStack.push_back("No info");
        }
      }
      else
      {
      }
    }
    free(undecoratedName);
  }

  /**
   * Dumps the call stack on the stream
   * @param stream is the stream to which the dump will be made
   */
  void dumpCallStack(std::ostream& stream)
  {
    stream << "**\nCallStack\n";

    for(std::list<std::string>::const_iterator it = callStack.begin(), it2 = fileStack.begin(); it != callStack.end() && it2 != fileStack.end(); ++it, ++it2)
    {
      stream << "- " << *it << std::endl;
      stream << "   " << *it2 << std::endl;
    }
    stream << "**" << std::endl;
  }

  /// The exception filter that will display the stack
  static LONG WINAPI UnhandledExceptionFilter(PEXCEPTION_POINTERS pExceptionInfo)
  {
    BacktraceImpl backtrace(pExceptionInfo);
    backtrace.dump(string);
    return EXCEPTION_CONTINUE_SEARCH;
  }

  /**
   * Backtrace constructor
   * @param pExceptionInfo is a pointer to a structure containing exception infos
   */
  BacktraceImpl(PEXCEPTION_POINTERS pExceptionInfo)
  {
    SymSetOptions(SYMOPT_UNDNAME|SYMOPT_DEFERRED_LOADS|SYMOPT_LOAD_LINES);
    hProcess = GetCurrentProcess();

    if(SymInitialize(hProcess, NULL, TRUE))
    {
      LoadCallStack(pExceptionInfo);
      EnumerateLoadedModules64(hProcess, EnumerateLoadedModules, &loadedModules);
      
      SymCleanup(hProcess);
    }
  }

  /**
   * Dumps the backtrace on a stream
   * @param stream is the stream on which the dump will be made
   */
  void dump(const std::string& string)
  {
    std::ofstream stream(string.c_str());
    dumpLoadedModules(stream);
    dumpCallStack(stream);
  }

  ~BacktraceImpl()
  {
  }
};

void setExceptionHandler(const std::string& string)
{
  BacktraceImpl::string = string;
  SetUnhandledExceptionFilter(BacktraceImpl::UnhandledExceptionFilter);
}

std::string BacktraceImpl::string;
