/** \file test_functions.h
 * Some test functions and variables.
 * This files includes the prototypes of some conversion functions, the declaration of assertions functions and macros
 */

#ifndef CLASS_TEST_H
#define CLASS_TEST_H

#include <QtCore/QString>
#include <sstream>

#ifdef WIN32
# ifdef testerlib_EXPORTS
#  define export_tester_dll
//__declspec(dllexport)
# else
#  define export_tester_dll
//__declspec(dllimport)
# endif
#else
#define export_tester_dll
#endif

#include <boost/utility.hpp>
#include <boost/mpl/not.hpp>
#include <boost/type_traits.hpp>

namespace matrix
{
  template<class DataType, unsigned int Height, unsigned int Width>
      struct Matrix;
}

namespace Tester
{
  /// Class trown when an assert is false
  struct export_tester_dll Thrown
  {
    Thrown(const QString& title, const QString& explanation, const QString& line, const QString& function, const QString& file)
      :title(title), explanation(explanation), line(line), function(function), file(file)
    {
    }
    /// Title of the exception
    QString title;
    /// Explanation of the exception
    QString explanation;
    /// Line in the file where the exception was raised
    QString line;
    /// Function in the file where the exception was raised
    QString function;
    /// File where the exception was raised
    QString file;
  };

  /// Creates a QString from an arithmetic value
  template<class ArithmeticType>
  typename boost::enable_if<boost::is_arithmetic<ArithmeticType>, QString>::type toQString(ArithmeticType number)
  {
    return QString::number(number);
  }
  /// Creates a QString from a matrix
  template<class MatrixType>
  typename boost::enable_if<boost::mpl::not_<boost::is_arithmetic<MatrixType> >, QString>::type toQString(const MatrixType& matrix)
  {
    std::stringstream oss;
    oss << matrix;
    return QString::fromStdString(oss.str());
  }

  /// Assertion for equality
  template<class T1, class T2>
  void assertEqual(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(rhs == lhs)
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert Equal (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Equal (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" Not Equal to ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for no equality
  template<class T1, class T2>
  void assertNotEqual(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(!(rhs == lhs))
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert NotEqual (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Equal (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" different from ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for difference
  template<class T1, class T2>
  void assertDifferent(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(rhs != lhs)
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert Different (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Different (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" Not different from ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for no difference
  template<class T1, class T2>
  void assertNotDifferent(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(!(rhs != lhs))
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert NotDifferent (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Different (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" Equal from ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for inferiority
  template<class T1, class T2>
  void assertLess(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(rhs < lhs)
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert Less (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Less (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" Not less than ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for no inferiority
  template<class T1, class T2>
  void assertNotLess(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(rhs < lhs)
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert NotLess (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Less (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" Greater than ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for superiority
  template<class T1, class T2>
  void assertGreater(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(rhs > lhs)
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert Great (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Great (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" Not greater than ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for no superiority
  template<class T1, class T2>
  void assertNotGreater(const T1& rhs, const T2& lhs, unsigned int line, const char* function, const char* file)
  {
    if(!(rhs > lhs))
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert NotGreat (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert Great (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" Less than ") + toQString(lhs), QString::number(line), function, file));
    }
  }

  /// Assertion for true
  template<class T1>
  void assertTrue(const T1& rhs, unsigned int line, const char* function, const char* file)
  {
    if(rhs)
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert True (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert True (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" not True"), QString::number(line), function, file));
    }
  }

  /// Assertion for false
  template<class T1>
  void assertFalse(const T1& rhs, unsigned int line, const char* function, const char* file)
  {
    if(!rhs)
    {
    }
    else
    {
      throw(Thrown(QString::fromUtf8("Failure in assert False (l") + toQString(line) + (")"), QString::fromUtf8("Failure in assert False (l") + toQString(line) + ("): ") + toQString(rhs) + QString::fromUtf8(" not False"), QString::number(line), function, file));
    }
  }

  /// Assert that only one kind of exception is thrown
  template<class Exception>
  void assertNoThrow(void (FunctionPtr)(), unsigned int line, const char* function, const char* file)
  {
    try
    {
      FunctionPtr();
    }
    catch(const Exception& exception)
    {
    }
  }
};

#define assertEqual(a, b) Tester::assertEqual(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertNotEqual(a, b) Tester::assertNotEqual(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertDifferent(a, b) Tester::assertDifferent(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertNotDifferent(a, b) Tester::assertNotDifferent(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertLess(a, b) Tester::assertLess(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertNotLess(a, b) Tester::assertNotLess(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertGreater(a, b) Tester::assertGreater(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertNotGreater(a, b) Tester::assertNotGreater(a, b, __LINE__, __FUNCTION__, __FILE__)
#define assertTrue(a) Tester::assertTrue(a, __LINE__, __FUNCTION__, __FILE__)
#define assertFalse(a) Tester::assertFalse(a, __LINE__, __FUNCTION__, __FILE__)
#define assertNoThrow(a, b) Tester::assertNoThrow<a>(b, __LINE__, __FUNCTION__, __FILE__)
#define thrown(a, b) Tester::Thrown(a, b, QString::number(__LINE__), __FUNCTION__, __FILE__)

#endif
