/** \file test_gui.h
 * The GUI for the tester.
 * This file describes the Qt GUI class for the tester.
 */

#ifndef TEST_GUI_H
#define TEST_GUI_H

#include <QtGui/QMainWindow>

class QVariant;
class QAction;
class QApplication;
class QButtonGroup;
class QFrame;
class QGridLayout;
class QHBoxLayout;
class QLabel;
class QListView;
class QListWidget;
class QListWidgetItem;
class QProgressBar;
class QPushButton;
class QTextBrowser;
class QTextEdit;
class QTreeWidget;
class QTreeWidgetItem;
class QVBoxLayout;
class QWidget;

struct TestFunctionStruct;
#include "test_functions.h"
#include "test_register_function.h"

/// The namespace which the tester is in
namespace Tester
{
  /// The GUI for the different tests
  class export_tester_dll TestMainWindow : public QMainWindow
  {
    Q_OBJECT
  public:
    /** The constructor
     * @param title is the title of the window
     */
    TestMainWindow(const QString title);

    unsigned int tests;
    unsigned int size_tests;
    unsigned int errors;
    unsigned int failures;

  private slots:
    /// Runs the selected tests
    void runTest();
    /// Changed selected test in the tests tree
    void treeTestChanged(QTreeWidgetItem* current, QTreeWidgetItem* previous);
    /// Changed selected failure/error
    void testChanged(QListWidgetItem* item);

  private:
    /// Initialises GUI/non-GUI elements of the Tester
    void clear();
    /// Creates the actions for each event
    void createActions();
    /// Fills the category box
    void fillTree();
    /**
     * Add a subtree to a node
     * @param node is the node to which the elemtn will be added
     * @param subTree is the subtree
     */
    void addToTree(QTreeWidgetItem* node, std::pair<const QString, TestFunctionStruct*>& subTree);
    /// Adds a failure
    /**
     * @param failure is a description of the failure
     */
    void addFailure(const Tester::Thrown& failure);
    /**
     * Adds an error
     * @param function is the name of the test function in which the error was detected
     * @param comment is the comment of the error
     */
    void addError(const QString& function, const QString comment);
    /**
     * Adds an error
     * @param function is the name of the test function in which the error was detected
     */
    void addError(const QString& function);
    /// Updates the ProgressBar
    void updateProgressBar();
    /// Vector of failures and error reports
    std::vector<QString> errorsFailuresReport;
    /**
     * Returns a QString describing the path to take to go from root to element
     * @param element is the element to find
     * @return the QString with the path
     */
    QString path(QTreeWidgetItem* element);
    /**
     * Returns a pointer to the test structure referenced by element
     * @param element is the element to find
     * @return the test structure associated to element
     */
    TestFunctionStruct* getSelectedTestStruct(QTreeWidgetItem* element);

  public:
    QLabel *nb_errors;
    QLabel *nb_tests;
    QLabel *nb_failures;
    QProgressBar *progressBar;
    QListWidget *listView;
    QTextEdit *textBrowser;
    QTreeWidget* tree;

  private:
    QIcon* iconFailure;
    QIcon* iconError;
    QWidget *centralWidget;
    QGridLayout *gridLayout;
    QFrame *displayInfoFrame;
    QVBoxLayout *verticalDisplayLayout;
    QVBoxLayout *verticalButtonDisplayLayout;
    QHBoxLayout *horizontalDisplayLayout;
    QPushButton *runButton;
    QFrame *frame;
    QGridLayout *gridLayout1;
    QGridLayout *textLayout;
    QLabel *labelFailures;
    QLabel *labelTests;
    QLabel *labelErrors;

    /// Sets up the GUI
    void setupUi(QMainWindow *MainWindow);
    /// Puts the initial values in the GUI
    void retranslateUi(QMainWindow *MainWindow);
  };
}

#endif // TEST_GUI_H
