/**
 * \file test_register_function.h
 * This contains structs to register the test functions
 */

#ifndef TEST_REGISTER_FUNCTION
#define TEST_REGISTER_FUNCTION

/* Registration Code */
#include <map>
#include <list>
#include <algorithm>
#include <stdexcept>

#include <QtCore/QString>
#include <QtCore/QStringList>

#include <testerlib/test_functions.h>

/* Registration mecanism */

struct TestComposite;
struct TestFunctionStruct;

/// The manager of all tests
struct export_tester_dll TestManager
{
protected:
  /// Protected constructor, only accessible by static function
  TestManager();

  /// The root of the test tree
  TestComposite* tree;

public:
  /// Creates an instance if not already created
  static TestManager& getInstance();

  /**
   * Gets the root of the tests tree
   * @return a pointer to the TestCompsite that is root
   */
  TestComposite* getRoot()
  {
    return tree;
  }
  /**
   * Gets a TestFunctionStruct from a QStringList
   * @param path is a list of QStrings
   * @return a TestFunctionStruct*
   */
  TestFunctionStruct* getTestFunctionStructFromQStringList(QStringList path);
  /**
   * Add a function to the test manager
   * @param name is the name of the test
   * @param f is a pointer to the test
   */
  void registerFunctions(const QString& name, TestFunctionStruct* f);
};

/// The parent test structure
struct export_tester_dll TestFunctionStruct
{
  /// Default constructor
  TestFunctionStruct(QString function)
    :functionName(function)
  {};
  /// Name of the structure
  QString functionName;

  /**
   * Returns the number of tests it contains
   * @return the number of tests
   */
  virtual unsigned int nbTests() = 0;
  /// Executes the test function
  virtual void operator()() = 0;
  /// Adds a element to the tree
  virtual void addTest(QStringList, TestFunctionStruct*) = 0;
  /**
   * Tells if node has leaves
   * @return false by default
   */
  virtual bool hasLeaves() const {return false;}
  /**
   * Get leaves
   * @return a pointer to leaves
   */
  virtual std::map<QString, TestFunctionStruct*>& getLeaves()
  {
    throw std::runtime_error("Does not have leaves");
  }
  /**
   * Returns all available tests
   * @return a list of tests to process
   */
  virtual std::list<TestFunctionStruct*> getRealTests() = 0;
  /// Destructor
  virtual ~TestFunctionStruct(){};
};

/// The real test suite
template<void fonction()>
struct TestSuite: public TestFunctionStruct
{
  /**
   * Constructor
   * @param name is the name of the test
   */
  TestSuite(QString name)
  :TestFunctionStruct(name)
  {
    TestManager::getInstance().registerFunctions(name, this);
  }

  /// Executes the test
  void operator()()
  {
    fonction();
  }

  /**
   * Returns all available tests
   * @return a list of tests to process
   */
  virtual std::list<TestFunctionStruct*> getRealTests()
  {
    std::list<TestFunctionStruct*> list;
    list.push_back(this);
    return list;
  }

  /**
   * Returns the number of tests it contains
   * @return the number of tests
   */
  unsigned int nbTests() {return 1;};

private:
  /// This function should not be called on a leaf
  void addTest(QStringList, TestFunctionStruct*)
  {}
};

/// The composite TestFunctionStruct
struct export_tester_dll TestComposite: public TestFunctionStruct
{
  /// The simple constructor
  TestComposite(QString name)
  :TestFunctionStruct(name), nbOfTests(0)
  {
  }

  /// Executes the test
  void operator()()
  {
  }

  /**
   * Returns all available tests
   * @return a list of tests to process
   */
  virtual std::list<TestFunctionStruct*> getRealTests()
  {
    std::list<TestFunctionStruct*> allTests;
    for(std::map<QString, TestFunctionStruct*>::iterator subNode = subTree.begin(); subNode != subTree.end(); ++subNode)
    {
      if(subNode->second->hasLeaves())
      {
        std::list<TestFunctionStruct*> nodeTests = subNode->second->getRealTests();
        std::copy(nodeTests.begin(), nodeTests.end(), std::back_inserter(allTests));
      }
      else
        allTests.push_back(subNode->second);
    }
    return allTests;
  }

  /**
   * Adds a new test to the tree
   * @param path is the path of the test in the tree
   * @param test is the test that must be saved
   */
  void addTest(QStringList path, TestFunctionStruct* test)
  {
    ++nbOfTests;
    if(path.size() == 1)
    {
      subTree[path[0]] = test;
    }
    else
    {
      if(subTree.find(path[0]) == subTree.end())
      {
        subTree[path[0]] = new TestComposite(path[0]);
      }
      std::map<QString, TestFunctionStruct*>::iterator element = subTree.find(path[0]);
      path.pop_front();
      element->second->addTest(path, test);
    }
  }

  /**
   * Indicates that the node has leaves
   * @return true
   */
  bool hasLeaves() const
  {
    return true;
  }

  /**
   * Get leaves
   * @return a pointer to leaves
   */
  std::map<QString, TestFunctionStruct*>& getLeaves()
  {
    return subTree;
  }

  /**
   * Returns the number of tests it contains
   * @return the number of tests
   */
  unsigned int nbTests() {return nbOfTests;};

private:
  unsigned int nbOfTests;
  /// The subtree
  std::map<QString, TestFunctionStruct*> subTree;
};
#endif
